/*
** main.c
**
** TicTacToe
**
** Written by Simion Andrei 331CB - 2017
** The Faculty of Automatic Control and Computer Science - UPB
*/

#include <avr/io.h>
#include <avr/interrupt.h>
#include <util/delay.h>
#include <avr/pgmspace.h>
#include <stdio.h>
#include <stdlib.h>
#include "lcd.h"

#define BUFF_X_MAX      (LCD_X_RES)
#define BUFF_Y_MAX      (LCD_Y_RES / 8)

#define M 0x7FFFFFFF

#define NO_LINES 100

typedef enum {START, RUNNING, END} STATE;
typedef enum {RIGHT = PB2, DOWN = PB1, UP = PB0, LEFT = PB3} BUTTON;
typedef enum {BLACK, WHITE} COLOR;
typedef enum {FALSE = 0, TRUE = 1} BOOL;
typedef enum {PLAYER1 = 1, PLAYER2 = 2, DRAW = 0, NONE = -1} PLAYER;

typedef struct
{
    STATE state;
    uint8_t lcd_buff[BUFF_Y_MAX][BUFF_X_MAX];
} GAME;

GAME game;
uint64_t rnd = 0;

//the working table
uint8_t table[3][3];

//player to move
PLAYER player = PLAYER1;

//current box, can't be bigger than 2(only 3 boxes)
uint8_t crt_x = 0;
uint8_t crt_y = 0;

static void delay(uint8_t num)
{
	uint8_t i;

	for (i = 0; i < num; i++)
		_delay_ms(500);
}


static BOOL pressed_button(BUTTON btn)
{
	return (PINB & (1 << btn)) == 0 ? TRUE : FALSE;
}


static void set_pixel(uint8_t x, uint8_t y, COLOR color)
{
	uint8_t yy = LCD_Y_RES - 1 - y;
	uint8_t row = yy / 8;
	uint8_t val = game.lcd_buff[row][x];

	switch (color)
	{
	case BLACK:
		val |= (1 << (yy % 8));
		break;
	case WHITE:
		val &= ~(1 << (yy % 8));
		break;
	default:
		return;
	}

	game.lcd_buff[row][x] = val;

	lcd_send(0x80 | x, LCD_CMD);    // column
	lcd_send(0x40 | row, LCD_CMD);  // row

	lcd_send(val, LCD_DATA);
}


static void clear_screen()
{
	uint8_t x, y;

	for (x = 0; x < LCD_X_RES; x++)
	{
		for (y = 0; y < LCD_Y_RES; y++)
		{
			set_pixel(x, y, WHITE);
		}
	}
}

static void draw_start()
{
	clear_screen();
	lcd_goto_xy(2, 1);
	lcd_str("Andrei Simion");
	lcd_goto_xy(4, 3);
	lcd_str("TicTacToe");
	lcd_goto_xy(1, 5);
	lcd_str("Click a Button");
	lcd_goto_xy(4, 6);
	lcd_str("to Start");

}

static void init_game()
{
	DDRB = 0x00;
	PORTB = 0xFF;

	MCUCR = _BV(JTD);
	MCUCR = _BV(JTD);

	game.state = START;
}

static void draw_table()
{
	uint8_t i;

	clear_screen();

	//draw horizontal lines
	for (i = 0; i < LCD_X_RES; i++)
	{
		set_pixel(i, LCD_Y_RES / 3, BLACK);
		set_pixel(i, (LCD_Y_RES / 3) * 2, BLACK);
	}
	for (i = 0; i < LCD_Y_RES; i++)
	{
		set_pixel(LCD_X_RES/3, i, BLACK);
		set_pixel((LCD_X_RES / 3) * 2, i, BLACK);
	}
}

static void blink_box()
{
	uint8_t x, y;
	uint8_t x_size = (LCD_X_RES / 3);
	uint8_t y_size = (LCD_Y_RES / 3);
	if (table[crt_x][crt_y] != 0)
	{
		return;
	}

	for (x = crt_x * x_size + 1; x < (crt_x + 1) * x_size - 1; x++)
	{
		for (y = crt_y * y_size + 1; y < (crt_y + 1) * y_size - 1; y++)
		{
			if (rnd % 10 == 0)
			{
				set_pixel(x, y, BLACK);
			}
			else
			{
				set_pixel(x, y, WHITE);
			}
		}
	}
}

static void clear_box()
{
	uint8_t x, y;
	uint8_t x_size = (LCD_X_RES / 3);
	uint8_t y_size = (LCD_Y_RES / 3);

	for (x = crt_x * x_size + 1; x < (crt_x + 1) * x_size - 1; x++)
	{
		for (y = crt_y * y_size + 1; y < (crt_y + 1) * y_size - 1; y++)
		{
			set_pixel(x, y, WHITE);
		}
	}
}

static void draw_o()
{
	uint8_t x, y;
	uint8_t x_size = (LCD_X_RES / 3);
	uint8_t y_size = (LCD_Y_RES / 3);

	clear_box();

	for (x = crt_x * x_size + 1; x < (crt_x + 1) * x_size - 1; x++)
	{
		for (y = crt_y * y_size + 1; y < (crt_y + 1) * y_size - 1; y++)
		{
			if (x >= crt_x * x_size + 6	&& x <= (crt_x + 1) * x_size - 6
					&& ((y == crt_y * y_size + 4)  || (y == (crt_y + 1) * y_size - 4)))
			{
				set_pixel(x, y, BLACK);
			}
			if (y >= crt_y * y_size + 4 && y <= (crt_y + 1) * y_size - 4
				&& ((x == crt_x * x_size + 6) || (x == (crt_x + 1) * x_size - 6)))
			{
				set_pixel(x, y, BLACK);
			}
		}
	}
}

static void draw_x()
{	
	uint8_t x, y;
	uint8_t x_size = (LCD_X_RES / 3);
	uint8_t y_size = (LCD_Y_RES / 3);
	
	clear_box();

	for (x = crt_x * x_size + 1; x < (crt_x + 1) * x_size - 1; x++)
	{
		for (y = crt_y * y_size + 1; y < (crt_y + 1) * y_size - 1; y++)
		{
			if (x >= crt_x * x_size + 6 && x <= (crt_x + 1) * x_size - 6
				&& y >= crt_y * y_size + 4 && y <= (crt_y + 1) * y_size - 4)
			{
				set_pixel(x, y, BLACK);
			}
		}
	}
	
}

static void move_right()
{
	if (table[crt_x][crt_y] == 0)
	{
		clear_box();
	}
	if (crt_x < 2)
	{
		crt_x++;
		return;
	}
	if (crt_x == 2 && crt_y < 2)
	{
		crt_y++;
		crt_x = 0;
		return;
	}
	if (crt_x == 2 && crt_y == 2)
	{
		crt_x = 0;
		crt_y = 0;
		return;
	}
}

static void move_left()
{
	if (table[crt_x][crt_y] == 0)
	{
		clear_box();
	}
	if (crt_x > 0)
	{
		crt_x--;
		return;
	}
	if (crt_x == 0 && crt_y > 0)
	{
		crt_y--;
		crt_x = 2;
		return;
	}
	if (crt_x == 0 && crt_y == 0)
	{
		crt_x = 2;
		crt_y = 2;
		return;
	}
}

static PLAYER check_winner()
{
	uint8_t i, j;

	BOOL ended = TRUE;
	for (i = 0; i < 3; i++)
	{
		for (j = 0; j < 3; j++)
		{
			if (table[i][j] == 0)
			{
				ended = FALSE;
			}
		}
	}
	if (ended == TRUE)
	{
		return DRAW;
	}

	for (i = 0; i < 3; i++)
	{
		if (table[i][0] == table[i][1] && table[i][0] == table[i][2] && table[i][0] != 0)
		{
			return table[i][0];
		}
		if (table[0][i] == table[1][i] && table[0][i] == table[2][i] && table[0][i] != 0)
		{
			return table[0][i];
		}
	}
	if (((table[0][0] == table[1][1] && table[0][0] == table[2][2])
		|| (table[2][0] == table[1][1] && table[2][0] == table[0][2])) && table[1][1] != 0)
	{
		return table[1][1];
	}

	return NONE;
}

static void make_move()
{
	if (player == PLAYER1)
	{
		draw_x();
		table[crt_x][crt_y] = PLAYER1;
		player = PLAYER2;
	}
	else
	{
		draw_o();
		table[crt_x][crt_y] = PLAYER2;
		player = PLAYER1;
	}
	if (check_winner() != NONE)
	{
		clear_screen();
		game.state = END;
		return;
	}
	while (table[crt_x][crt_y] != 0)
		move_right();
}

static void draw_end()
{
	lcd_goto_xy(4, 3);
	lcd_str("GAME OVER");

	lcd_goto_xy(1, 5);

	PLAYER winner = check_winner();
	switch (winner)
	{
	case DRAW:
		lcd_goto_xy(6, 5);
		lcd_str("DRAW");
		break;
	case PLAYER1:
		lcd_goto_xy(2, 5);
		lcd_str("PLAYER 1 WINS");
		break;
	case PLAYER2:
		lcd_goto_xy(2, 5);
		lcd_str("PLAYER 2 WINS");
		break;
	default:
		break;
	}
}

int main(void)
{
    lcd_init();
    lcd_contrast(0x40);

    draw_start();
    delay(1);
	init_game();

	while (TRUE)
	{
		rnd++;
		switch (game.state)
		{
		case START:
			if (pressed_button(UP)
				|| pressed_button(DOWN)
				|| pressed_button(LEFT)
				|| pressed_button(RIGHT))
			{
				draw_table();
				game.state = RUNNING;
			}
			delay(1);
			break;
		case RUNNING:
			blink_box();
			if (pressed_button(DOWN))
			{
				make_move();
			}
			else if (pressed_button(LEFT))
			{
				do
					move_left();
				while(table[crt_x][crt_y] != 0);
			}
			else if (pressed_button(RIGHT))
			{
				do
					move_right();
				while (table[crt_x][crt_y] != 0);
			}
			delay(1);
			break;
		case END:
			draw_end();
			break;
		default:
			break;
		}
	}

    return 0;
}